# -*- coding: utf-8 -*-
"""
    sphinxcontrib.aafig
    ~~~~~~~~~~~~~~~~~~~

    Allow embeded ASCII art to be rendered as nice looking images
    using the aafigure reStructuredText extension.

    See the README file for details.

    :author: Leandro Lucarella <llucax@gmail.com>
    :license: BOLA
"""

# original license, copied from LICENSE file:
# # I don't like licenses, because I don't like having to worry about all this
# # legal stuff just for a simple piece of software I don't really mind anyone
# # using. But I also believe that it's important that people share and give back;
# # so I'm placing this work under the following license.
# #
# #
# # BOLA - Buena Onda License Agreement (v1.0)
# # ------------------------------------------
# #
# # This work is provided 'as-is', without any express or implied warranty. In no
# # event will the authors be held liable for any damages arising from the use of
# # this work.
# #
# # To all effects and purposes, this work is to be considered Public Domain.
# #
# #
# # However, if you want to be "buena onda", you should:
# #
# # 1. Not take credit for it, and give proper recognition to the authors.
# # 2. Share your modifications, so everybody benefits from them.
# # 3. Do something nice for the authors.
# # 4. Help someone who needs it.
# # 5. Don't waste. Anything, but specially energy that comes from natural
# #    non-renewable resources.
# # 6. Be tolerant. Everything that's good in nature comes from cooperation.

import posixpath
from os import path
try:
    from hashlib import sha1 as sha
except ImportError:
    from sha import sha

from docutils import nodes
from docutils.parsers.rst.directives import images, nonnegative_int, flag

from sphinx.errors import SphinxError
from sphinx.util import ensuredir, relative_uri

from . import aafigure


DEFAULT_FORMATS = dict(html='svg', latex='pdf', text=None)


def merge_dict(dst, src):
    for (k, v) in src.items():
        if k not in dst:
            dst[k] = v
    return dst


def get_basename(text, options, prefix='aafig'):
    options = options.copy()
    if 'format' in options:
        del options['format']
    hashkey = (text + str(options)).encode('utf-8')
    id = sha(hashkey).hexdigest()
    return '{}-{}'.format(prefix, id)


class AafigError(SphinxError):
    category = 'aafig error'


class AafigDirective(images.Image):
    """
    Directive to insert an ASCII art figure to be rendered by aafigure.
    """
    has_content = True
    required_arguments = 0
    own_option_spec = dict(
        line_width=float,
        background=str,
        foreground=str,
        fill=str,
        aspect=nonnegative_int,
        scale=nonnegative_int,
        textual=flag,
        proportional=flag,
        rounded=flag,
    )
    option_spec = images.Image.option_spec.copy()
    option_spec.update(own_option_spec)

    def run(self):
        aafig_options = dict()
        own_options_keys = self.own_option_spec.keys()
        for (k, v) in list(self.options.items()):
            if k in own_options_keys:
                # convert flags to booleans
                if v is None:
                    v = True
                # convert percentage to float
                if k == 'scale' or k == 'aspect':
                    v = float(v) / 100.0
                aafig_options[k] = v
                del self.options[k]
        self.arguments = ['']
        (image_node,) = images.Image.run(self)
        if isinstance(image_node, nodes.system_message):
            return [image_node]
        text = '\n'.join(self.content)
        image_node.aafig = dict(options=aafig_options, text=text)
        return [image_node]


def render_aafig_images(app, doctree):
    format_map = app.builder.config.aafig_format
    merge_dict(format_map, DEFAULT_FORMATS)
    for img in doctree.traverse(nodes.image):
        if not hasattr(img, 'aafig'):
            continue
        options = img.aafig['options']
        text = img.aafig['text']
        format = app.builder.format
        merge_dict(options, app.builder.config.aafig_default_options)
        if format in format_map:
            options['format'] = format_map[format]
        else:
            app.builder.warn(
                'unsupported builder format "{}", please '
                'add a custom entry in aafig_format config option '
                'for this builder'.format(format))
            img.replace_self(nodes.literal_block(text, text))
            continue
        if options['format'] is None:
            img.replace_self(nodes.literal_block(text, text))
            continue
        try:
            fname, outfn, id, extra = render_aafigure(app, text, options)
        except AafigError as exc:
            app.builder.warn('aafigure error: {}'.format(exc))
            img.replace_self(nodes.literal_block(text, text))
            continue
        img['uri'] = fname
        # FIXME: find some way to avoid this hack in aafigure
        if extra:
            (width, height) = [x.split('"')[1] for x in extra.split()]
            if not img.has_key('width'):
                img['width'] = width
            if not img.has_key('height'):
                img['height'] = height


def render_aafigure(app, text, options):
    """
    Render an ASCII art figure into the requested format output file.
    """
    fname = get_basename(text, options)
    fname = '{}.{}'.format(get_basename(text, options), options['format'])
    if app.builder.format == 'html':
        # HTML
        imgpath = relative_uri(app.builder.env.docname, '_images')
        relfn = posixpath.join(imgpath, fname)
        outfn = path.join(app.builder.outdir, '_images', fname)
    else:
        # Non-HTML
        if app.builder.format != 'latex':
            app.builder.warn(
                'aafig: the builder format {} is not officially '
                'supported, aafigure images could not work. Please report '
                'problems and working builder to avoid this warning in '
                'the future'.format(app.builder.format))
        relfn = fname
        outfn = path.join(app.builder.outdir, fname)
    metadata_fname = '{}.aafig'.format(outfn)

    try:
        if path.isfile(outfn):
            extra = None
            if options['format'].lower() == 'svg':
                f = None
                try:
                    try:
                        f = open(metadata_fname, 'r')
                        extra = f.read()
                    except:
                        raise AafigError()
                finally:
                    if f is not None:
                        f.close()
            return relfn, outfn, id, extra
    except AafigError:
        pass

    ensuredir(path.dirname(outfn))

    try:
        (visitor, output) = aafigure.render(text, outfn, options)
        output.close()
    except aafigure.UnsupportedFormatError as e:
        raise AafigError(str(e))

    extra = None
    if options['format'].lower() == 'svg':
        extra = visitor.get_size_attrs()
        with open(metadata_fname, 'w') as f:
            f.write(extra)

    return relfn, outfn, id, extra


def setup(app):
    app.add_directive('aafig', AafigDirective)
    app.connect('doctree-read', render_aafig_images)
    app.add_config_value('aafig_format', DEFAULT_FORMATS, 'html')
    app.add_config_value('aafig_default_options', dict(), 'html')


# vim: set expandtab shiftwidth=4 softtabstop=4 :



#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2006 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
Bitmap renderer for the aafigure package, using the Python Imaging Library.
"""

import sys
from .error import UnsupportedFormatError
try:
    from PIL import Image, ImageDraw
except ImportError:
    raise UnsupportedFormatError('please install PIL or Pillow to get bitmap output support')
from . import PILhelper


class PILOutputVisitor:
    """Render a list of shapes as bitmap."""

    def __init__(self, options):
        self.options = options
        self.scale = options['scale'] * 8
        self.debug = options['debug']
        self.line_width = options['line_width']
        self.foreground = options['foreground']
        self.background = options['background']
        self.fillcolor = options['fill']

    def _num(self, number):
        return number * self.scale

    def visit_image(self, aa_image):
        """\
        Process the given ASCIIArtFigure and draw the shapes in
        the bitmap file
        """
        self.aa_image = aa_image        # save for later XXX not optimal to do it here
        self.width = (aa_image.width + 1) * aa_image.nominal_size * aa_image.aspect_ratio
        self.height = (aa_image.height + 1) * aa_image.nominal_size

        # if font is given explicit, use it instead of proportional flag
        font_size = int(self._num(self.aa_image.nominal_size * 1.1))
        if 'font' in self.options:
            self.font = PILhelper.font_by_name(self.options['font'], font_size)
        else:
            self.font = PILhelper.font_by_type(self.options['proportional'], font_size)
        if self.font is None:
            sys.stderr.write("WARNING: font not found, using PIL default font\n")

        self.image = Image.new(
            'RGB',
            (int(self._num(self.width)), int(self._num(self.height))),
            self.background
        )
        self.draw = ImageDraw.Draw(self.image)

        #~ if self.debug:
            #~ #draw a rectangle around entire image
            #~ self._rectangle(
                #~ 0,0,
                #~ aa_image.width, aa_image.height,
                #~ style = 'fill:none;',
            #~ )

        self.visit_shapes(aa_image.shapes)
        del self.draw
        file_type = self.options['format'].lower()
        if file_type == 'jpg':
            file_type = 'jpeg'  # alias
        try:
            if 'file_like' in self.options:
                self.image.save(self.options['file_like'], file_type)
        except KeyError:
            raise UnsupportedFormatError("PIL doesn't support image format {!r}".format(file_type))

    def visit_shapes(self, shapes):
        for shape in shapes:
            shape_name = shape.__class__.__name__.lower()
            visitor_name = 'visit_{}'.format(shape_name)
            if hasattr(self, visitor_name):
                getattr(self, visitor_name)(shape)
            else:
                sys.stderr.write("WARNING: don't know how to handle shape {!r}\n".format(shape))

    def visit_group(self, group):
        self.visit_shapes(group.shapes)

    # - - - - - - drawing helpers - - - - - - -
    def _line(self, x1, y1, x2, y2):
        """Draw a line, coordinates given as four decimal numbers"""
        self.draw.line((self._num(x1), self._num(y1),
                        self._num(x2), self._num(y2)),
                       fill=self.foreground)  # self.line_width

    def _rectangle(self, x1, y1, x2, y2):
        """\
        Draw a rectangle, coordinates given as four decimal numbers.
        ``style`` is inserted in the SVG. It could be e.g. "fill:yellow"
        """
        self.draw.rectangle((self._num(x1), self._num(y1),
                             self._num(x2), self._num(y2)),
                            fill=self.fillcolor,
                            outline=self.foreground)  # self.line_width

    # - - - - - - visitor function for the different shape types - - - - - - -

    def visit_point(self, point):
        dotsize = 2
        self.draw.ellipse(
            (
                self._num(point.x) - dotsize, self._num(point.y) - dotsize,
                self._num(point.x) + dotsize, self._num(point.y) + dotsize
            ),
            fill=self.foreground
        )

    def visit_line(self, line):
        x1, x2 = line.start.x, line.end.x
        y1, y2 = line.start.y, line.end.y
        self._line(x1, y1, x2, y2)

    def visit_rectangle(self, rectangle):
        self._rectangle(
            rectangle.p1.x, rectangle.p1.y,
            rectangle.p2.x, rectangle.p2.y,
        )

    def visit_circle(self, circle):
        self.draw.ellipse(
            (
                self._num(circle.center.x - circle.radius), self._num(circle.center.y - circle.radius),
                self._num(circle.center.x + circle.radius), self._num(circle.center.y + circle.radius)
            ),
            fill=self.fillcolor,
            outline=self.foreground,
        )

    def visit_label(self, label):
        #  font-weight="bold"
        self.draw.text(
            (self._num(label.position.x), self._num(label.position.y - self.aa_image.nominal_size * 1.1)),
            label.text,
            fill=self.foreground,
            font=self.font
        )

    def _bezier(self, p1, c1, c2, p2, level=1):
        # de Casteljau's algorithm
        if self._num(p1.distance(p2)) <= 3:
            self._line(p1.x, p1.y, p2.x, p2.y)
        else:
            cmid = c1.midpoint(c2)
            lp1 = p1
            lc1 = p1.midpoint(c1)
            lc2 = lc1.midpoint(cmid)
            rp2 = p2
            rc2 = p2.midpoint(c2)
            rc1 = rc2.midpoint(cmid)
            lp2 = rc1.midpoint(lc2)
            rp1 = lp2
            self._bezier(lp1, lc1, lc2, lp2, level + 1)
            self._bezier(rp1, rc1, rc2, rp2, level + 1)

    def visit_arc(self, arc):
        p1, p2 = arc.start, arc.end
        c1 = arc.start_control_point()
        c2 = arc.end_control_point()
        self._bezier(p1, c1, c2, p2)



#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2006 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
Simple ASCII output of the rendered image.  Think of it as a low resolution
black and white image.
"""

import sys


class AsciiOutputVisitor:
    """\
    Render a list of shapes as ASCII art.
    Scaled, think of it as a low resolution black and white image.
    """

    def __init__(self, options):
        self.options = options
        self.image = {}
        self.scale = options['scale']

    def visit_image(self, aa_image):
        self.visit_shapes(aa_image.shapes)
        self.options['file_like'].write(self.create_image())

    def visit_shapes(self, shapes):
        for shape in shapes:
            shape_name = shape.__class__.__name__.lower()
            visitor_name = 'visit_{}'.format(shape_name)
            function = getattr(self, visitor_name, None)
            if function is not None:
                function(shape)
            else:
                sys.stderr.write("WARNING: don't know how to handle shape {!r}\n".format(shape))

    def visit_group(self, group):
        self.visit_shapes(group.shapes)

    def visit_point(self, point):
        self.image[point.x * self.scale, point.y * self.scale] = '#'

    def visit_line(self, line):
        x1, x2 = line.start.x * self.scale, line.end.x * self.scale
        y1, y2 = line.start.y * self.scale, line.end.y * self.scale
        if x1 > x2:
            x1, x2 = x2, x1
        if y1 > y2:
            y1, y2 = y2, y1
        dx = x2 - x1
        dy = y2 - y1
        if dx > dy:
            y = y1
            if dx:
                m = float(dy) / dx
            else:
                m = 0
            for x in range(int(x1), int(x2 + 1)):
                self.image[x, int(y)] = '#'
            y += m
        else:
            x = x1
            if dy:
                m = float(dx) / dy
            else:
                m = 0
            for y in range(int(y1), int(y2+1)):
                self.image[int(x), y] = '#'
            x += m

    def visit_rectangle(self, rectangle):
        x1, x2 = rectangle.p1.x * self.scale, rectangle.p2.x * self.scale
        y1, y2 = rectangle.p1.y * self.scale, rectangle.p2.y * self.scale
        if x1 > x2:
            x1, x2 = x2, x1
        if y1 > y2:
            y1, y2 = y2, y1
        for y in range(int(y1), int(y2)):
            for x in range(int(x1), int(x2)):
                self.image[x, y] = '#'

    def visit_label(self, label):
        x, y = int(label.position.x * self.scale), int(label.position.y * self.scale)
        for character in label.text:
            self.image[x, y] = character
            x += 1

    def __str__(self):
        return self.create_image()

    def create_image(self):
        """return a cropped image"""
        # find out size
        min_x = min_y = 1000000
        max_x = max_y = -1000000
        for x, y in self.image:
            min_x = min(min_x, x)
            max_x = max(max_x, x)
            min_y = min(min_y, y)
            max_y = max(max_y, y)
        # render image to lines of text, fill unused fields with a dot
        result = []
        for y in range(min_y, max_y+1):
            line = []
            for x in range(min_x, max_x+1):
                line.append(self.image.get((x, y), '.'))
            result.append(''.join(line))
        return u'{}\n'.format('\n'.join(result))



#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2010 Chris Liechti <cliechti@gmx.net>, Oliver Joos <oliver.joos@hispeed.ch>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
Helper objects for bitmap renderer of the aafigure package.
"""

import os
from PIL import ImageFont


# - - - - - - font helpers - - - - - - -

def _find_file(name, top_dir):
    """\
    Find a file by its name in a directory or sub-directories (recursively).
    Return absolute path of the file or None if not found.
    """
    for (dirpath, dirnames, filenames) in os.walk(top_dir):
        if name in filenames:
            return os.path.join(dirpath, name)
    return None


def font_by_name(name, size):
    """\
    Get a PIL ImageFont instance by font name and size. If name is not an
    absolute pathname, it is searched in the default font locations of the
    underlying OS. If not found, None is returned.
    """
    font = None
    try:
        font = ImageFont.truetype(name, size)
    except IOError:
        # PIL upto 1.1.7b1 only tries absolute paths for win32
        if os.name == 'posix':
            font_path = _find_file(name, '/usr/share/fonts')
            if font_path:
                try:
                    font = ImageFont.truetype(font_path, size)
                except IOError:
                    pass
    return font


def font_by_type(proportional, size):
    """\
    Get a PIL ImageFont instance by font type and size. If <proportional> is
    not True, a mono-spaced font is returned. If no suitable font is found,
    None is returned.
    """
    if proportional:
        font = font_by_name('LiberationSans-Regular.ttf', size)
        if font is None:
            font = font_by_name('Arial.ttf', size)
    else:
        font = font_by_name('LiberationMono-Regular.ttf', size)
        if font is None:
            font = font_by_name('Courier_New.ttf', size)
    return font



# Common shapes for the aafigure package.
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2009 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
# This intentionally is no doc comment to make it easier to include the module
# in Sphinx ``.. automodule::``
import math


def point(obj):
    """\
    return a Point instance.
    - if object is already a Point instance it's returned as is
    - complex numbers are converted to Points
    - a tuple with two elements (x,y)
    """
    if isinstance(obj, Point):
        return obj
    if type(obj) is complex:
        return Point(obj.real, obj.imag)
    if type(obj) is tuple and len(obj) == 2:
        return Point(obj[0], obj[1])
    raise ValueError('can not convert {!r} to a Point'.format(obj))


def group(list_of_shapes):
    """return a group if the number of shapes is greater than one"""
    if len(list_of_shapes) > 1:
        return [Group(list_of_shapes)]
    else:
        return list_of_shapes


class Point:
    """\
    A single point. This class primary use is to represent coordinates
    for the other shapes.
    """
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return 'Point({p.x!r}, {p.y!r})'.format(p=self)

    def distance(self, other):
        return math.sqrt((self.x - other.x) ** 2 +
                         (self.y - other.y) ** 2)

    def midpoint(self, other):
        return Point((self.x + other.x) / 2,
                     (self.y + other.y) / 2)


class Line:
    """Line with starting and ending point. Both ends can have arrows"""
    def __init__(self, start, end, thick=False):
        self.thick = thick
        self.start = point(start)
        self.end = point(end)

    def __repr__(self):
        return 'Line({l.start!r}, {l.end!r})'.format(l=self)


class Rectangle:
    """Rectangle with two edge coordinates."""
    def __init__(self, p1, p2):
        self.p1 = point(p1)
        self.p2 = point(p2)

    def __repr__(self):
        return 'Rectangle({r.p1!r}, {r.p2!r})'.format(r=self)


class Circle:
    """Circle with center coordinates and radius."""
    def __init__(self, center, radius):
        self.center = point(center)
        self.radius = radius

    def __repr__(self):
        return 'Circle({c.center!r}, {c.radius!r})'.format(c=self)


class Label:
    """A text label at a position"""
    def __init__(self, position, text):
        self.position = position
        self.text = text

    def __repr__(self):
        return 'Label({t.position!r}, {t.text!r})'.format(t=self)


class Group:
    """A group of shapes"""
    def __init__(self, shapes=None):
        if shapes is None:
            shapes = []
        self.shapes = shapes

    def __repr__(self):
        return 'Group({!r})'.format(self.shapes)


class Arc:
    """A smooth arc between two points"""
    def __init__(self, start, start_angle, end, end_angle, start_curve=True, end_curve=True):
        self.start = point(start)
        self.end = point(end)
        self.start_angle = start_angle
        self.end_angle = end_angle
        self.start_curve = start_curve
        self.end_curve = end_curve

    def __repr__(self):
        return 'Arc({a.start!r}, {a.start_angle!r}, ' \
               '{a.end!r}, {a.end_angle!r}, {a.start_curve!r}, ' \
               '{a.end_curve!r})'.format(a=self)

    def start_angle_rad(self):
        return self.start_angle * math.pi / 180

    def end_angle_rad(self):
        return self.end_angle * math.pi / 180

    def __tension(self):
        return self.start.distance(self.end) / 3

    # assumptions: x increases going right, y increases going down
    def start_control_point(self):
        if self.start_curve:
            dd = self.__tension()
            angle = self.start_angle_rad()
            return Point(self.start.x + dd * math.cos(angle),
                         self.start.y - dd * math.sin(angle))
        else:
            return self.start

    def end_control_point(self):
        if self.end_curve:
            dd = self.__tension()
            angle = self.end_angle_rad()
            return Point(self.end.x + dd * math.cos(angle),
                         self.end.y - dd * math.sin(angle))
        else:
            return self.end



#!/usr/bin/env python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2006-2011 Chris Liechti <cliechti@gmx.net> and the aafigure-team
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
ASCII art to image converter.

This is the main module that contains the parser.

See svg.py and aa.py for output modules, that can render the parsed structure.
"""
# import codecs
from .error import UnsupportedFormatError
from .shapes import Line, Point, Circle, Label, Arc, Rectangle, group  # point
from unicodedata import east_asian_width
import sys


try:
    basestring
except NameError:
    basestring = str

NOMINAL_SIZE = 2

CLASS_LINE = 'line'
CLASS_STRING = 'str'
CLASS_RECTANGLE = 'rect'
CLASS_JOIN = 'join'
CLASS_FIXED = 'fixed'

DEFAULT_OPTIONS = dict(
    background='#ffffff',
    foreground='#000000',
    line_width=2.0,
    scale=1.0,
    aspect=1.0,
    format='svg',
    debug=False,
    textual=False,
    textual_strict=False,
    proportional=False,
    encoding='utf-8',
    widechars='F,W',
    rounded=False,
)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


class AsciiArtImage:
    """\
    This class holds a ASCII art figure and has methods to parse it.
    The resulting list of shapes is also stored here.

    The image is parsed in 2 steps:

    1. horizontal string detection.
    2. generic shape detection.

    Each character that is used in a shape or string is tagged. So that
    further searches don't include it again (e.g. text in a string touching
    a fill), respectively can use it correctly (e.g. join characters when
    two or more lines hit).
    """

    QUOTATION_CHARACTERS = list('"\'`')

    def __init__(self, text, options=None):
        """Take a ASCII art figure and store it, prepare for ``recognize``"""
        if options is None: options = {}
        self.aspect_ratio = float(options.get('aspect_ratio', 1))
        self.textual = options.get('textual', False)
        self.textual_strict = options.get('textual_strict', False)
        self.rounded = options.get('rounded', False)
        # XXX TODO tab expansion
        # detect size of input image, store as list of lines
        self.image = []
        max_x = 0
        y = 0
        # define character widths map
        charwidths = {}
        for key in ['F', 'H', 'W', 'Na', 'A', 'N']:
            if key in options.get('widechars', 'F,W').split(','):
                charwidths[key] = 2
            else:
                charwidths[key] = 1
        for line in text.splitlines():
            # extend length by 1 for each wide glyph
            line_len = sum(charwidths[east_asian_width(c)] for c in line)
            max_x = max(max_x, line_len)
            # pad a space for each wide glyph
            padded_line = ''.join(c + ' ' * (charwidths[east_asian_width(c)] - 1) for c in line)
            self.image.append(padded_line)
            y += 1
        self.width = max_x
        self.height = y
        # make sure it's rectangular (extend short lines to max width)
        for y, line in enumerate(self.image):
            if len(line) < max_x:
                self.image[y] = line + ' ' * (max_x - len(line))
        # initialize other data structures
        self.classification = [[None] * self.width for y in range(self.height)]
        self.shapes = []
        self.nominal_size = NOMINAL_SIZE

    def __str__(self):
        """Return the original image"""
        return '\n'.join([self.image[y] for y in range(self.height)])

    def get(self, x, y):
        """\
        Get character from image. Gives no error for access out of
        bounds, just returns a space. This simplifies the scanner
        functions.
        """
        if 0 <= x < self.width and 0 <= y < self.height:
            return self.image[y][x]
        else:
            return ' '

    def tag(self, coordinates, classification):
        """Tag coordinates as used, store classification"""
        for x, y in coordinates:
            self.classification[y][x] = classification

    def cls(self, x, y):
        """Get tag at coordinate"""
        try:
            return self.classification[y][x]
        except IndexError:
            return 'outside'

    # Coordinate conversion and shifting
    def left(self, x):
        return x * NOMINAL_SIZE * self.aspect_ratio

    def hcenter(self, x):
        return (x + 0.5) * NOMINAL_SIZE * self.aspect_ratio

    def right(self, x):
        return (x + 1)*NOMINAL_SIZE * self.aspect_ratio

    def top(self, y):
        return y * NOMINAL_SIZE

    def vcenter(self, y):
        return (y + 0.5) * NOMINAL_SIZE

    def bottom(self, y):
        return (y + 1) * NOMINAL_SIZE

    def recognize(self):
        """\
        Try to convert ASCII art to vector graphics. The result is stored in
        ``self.shapes``.
        """
        # XXX search for symbols
        #~ #search for long strings
        #~ for y in range(self.height):
            #~ for x in range(self.width):
                #~ character = self.image[y][x]
                #~ if self.classification[y][x] is None:
                    #~ if character.isalnum():
                        #~ self.shapes.extend(
                            #~ self._follow_horizontal_string(x, y)
                        #~ )
        # search for quoted texts
        for y in range(self.height):
            for x in range(self.width):
                #if not yet classified, check for a line
                character = self.image[y][x]
                if character in self.QUOTATION_CHARACTERS and self.classification[y][x] is None:
                    self.shapes.extend(
                        self._follow_horizontal_string(x, y, quoted=True))

        # search for standard shapes
        for y in range(self.height):
            for x in range(self.width):
                # if not yet classified, check for a line
                character = self.image[y][x]
                if self.classification[y][x] is None:
                    if character == '-':
                        self.shapes.extend(self._follow_horizontal_line(x, y))
                    elif character == '|':
                        self.shapes.extend(self._follow_vertical_line(x, y))
                    elif character == '_':
                        self.shapes.extend(self._follow_lower_horizontal_line(x, y))
                    elif character == '~':
                        self.shapes.extend(self._follow_upper_horizontal_line(x, y))
                    elif character == '=':
                        self.shapes.extend(self._follow_horizontal_line(x, y, thick=True))
                    elif character in '\\/':
                        self.shapes.extend(self._follow_rounded_edge(x, y))
                    elif character == '+':
                        self.shapes.extend(self._plus_joiner(x, y))
                    elif character in self.FIXED_CHARACTERS:
                        self.shapes.extend(self.get_fixed_character(character)(x, y))
                        self.tag([(x, y)], CLASS_FIXED)
                    elif not self.textual_strict and character in self.FILL_CHARACTERS:
                        if self.textual:
                            if self.get(x, y + 1) == character:
                                self.shapes.extend(self._follow_fill(character, x, y))
                        else:
                            if (self.get(x + 1, y) == character or self.get(x, y + 1) == character):
                                self.shapes.extend(self._follow_fill(character, x, y))

        # search for short strings too
        for y in range(self.height):
            for x in range(self.width):
                character = self.image[y][x]
                if self.classification[y][x] is None:
                    if character != ' ':
                        self.shapes.extend(self._follow_horizontal_string(x, y, accept_anything=True))

    # - - - - - - - - - helper function for some shapes - - - - - - - - -
    # Arrow drawing functions return the (new) starting point of the line and a
    # list of shapes that draw the arrow. The line itself is not included in
    # the list of shapes. The stating point is p1, possibly modified to match
    # the shape of the arrow head.
    #
    # Use complex numbers as 2D vectors as that means easy transformations like
    # scaling, rotation and translation

    # - - - - - - - - - arrows - - - - - - - - -
    def _standard_arrow(self, p1, p2):
        """\
        -->
        Return a possibly modified starting point and a list of shapes.
        """
        direction_vector = p1 - p2
        direction_vector /= abs(direction_vector)
        return p1, [
            Line(p1, p1-direction_vector * 1.5 + direction_vector * 0.5j),
            Line(p1, p1-direction_vector * 1.5 + direction_vector * -0.5j)
        ]

    def _reversed_arrow(self, p1, p2):
        """--<"""
        direction_vector = p1 - p2
        direction_vector /= abs(direction_vector)
        return p1-direction_vector * 2, [
            Line(p1-direction_vector * 2.0, p1+direction_vector * (-0.5 + 0.5j)),
            Line(p1-direction_vector * 2.0, p1+direction_vector * (-0.5 - 0.5j))
        ]

    def _circle_head(self, p1, p2, radius=0.5):
        """--o"""
        direction_vector = p1 - p2
        direction_vector /= abs(direction_vector)
        return p1-direction_vector, [Circle(p1 - direction_vector, radius)]

    def _large_circle_head(self, p1, p2):
        """--O"""
        return self._circle_head(p1, p2, radius=0.9)

    def _rectangular_head(self, p1, p2):
        """--#"""
        direction_vector = p1 - p2
        direction_vector /= abs(direction_vector)
        #~ return p1-direction_vector*1.414, [
            #~ Rectangle(p1-direction_vector-direction_vector*(0.707+0.707j),
                      #~ p1-direction_vector+direction_vector*(0.707+0.707j))
        #~ ]
        return p1-direction_vector * 1.707, [
            Line(p1 - direction_vector - direction_vector * (0.707 + 0.707j),
                 p1 - direction_vector - direction_vector * (0.707 - 0.707j)),
            Line(p1 - direction_vector + direction_vector * (0.707 + 0.707j),
                 p1 - direction_vector + direction_vector * (0.707 - 0.707j)),
            Line(p1 - direction_vector - direction_vector * (0.707 + 0.707j),
                 p1 - direction_vector + direction_vector * (0.707 - 0.707j)),
            Line(p1 - direction_vector - direction_vector * (0.707 - 0.707j),
                 p1 - direction_vector + direction_vector * (0.707 + 0.707j)),
        ]

    # the same character can mean a different thing, depending from where the
    # line is coming. this table maps line direction (dx,dy) and the arrow
    # character to a arrow drawing function
    ARROW_TYPES = [
        # chr  dx  dy  arrow type
        ('>',  1,  0, '_standard_arrow'),
        ('<', -1,  0, '_standard_arrow'),
        ('^',  0, -1, '_standard_arrow'),
        ('A',  0, -1, '_standard_arrow'),
        ('V',  0,  1, '_standard_arrow'),
        ('v',  0,  1, '_standard_arrow'),
        ('>', -1,  0, '_reversed_arrow'),
        ('<',  1,  0, '_reversed_arrow'),
        ('^',  0,  1, '_reversed_arrow'),
        ('V',  0, -1, '_reversed_arrow'),
        ('v',  0, -1, '_reversed_arrow'),
        ('o',  1,  0, '_circle_head'),
        ('o', -1,  0, '_circle_head'),
        ('o',  0, -1, '_circle_head'),
        ('o',  0,  1, '_circle_head'),
        ('O',  1,  0, '_large_circle_head'),
        ('O', -1,  0, '_large_circle_head'),
        ('O',  0, -1, '_large_circle_head'),
        ('O',  0,  1, '_large_circle_head'),
        ('#',  1,  0, '_rectangular_head'),
        ('#', -1,  0, '_rectangular_head'),
        ('#',  0, -1, '_rectangular_head'),
        ('#',  0,  1, '_rectangular_head'),
    ]

    ARROW_HEADS = list('<>AVv^oO#')

    def get_arrow(self, character, dx, dy):
        """Return arrow drawing function or None."""
        for head, ddx, ddy, function_name in self.ARROW_TYPES:
            if character == head and dx == ddx and dy == ddy:
                return getattr(self, function_name)

    # - - - - - - - - - fills - - - - - - - - -
    # Fill functions return a list of shapes. Each one if covering one cell
    # size.

    def _hatch_left(self, x, y):
        return self._n_hatch_diagonal(x, y, 1, True)

    def _hatch_right(self, x, y):
        return self._n_hatch_diagonal(x, y, 1, False)

    def _cross_hatch(self, x, y):
        return self._n_hatch_diagonal(x, y, 1, True) + \
               self._n_hatch_diagonal(x, y, 1, False)

    def _double_hatch_left(self, x, y):
        return self._n_hatch_diagonal(x, y, 2, True)

    def _double_hatch_right(self, x, y):
        return self._n_hatch_diagonal(x, y, 2, False)

    def _double_cross_hatch(self, x, y):
        return self._n_hatch_diagonal(x, y, 2, True) + \
               self._n_hatch_diagonal(x, y, 2, False)

    def _triple_hatch_left(self, x, y):
        return self._n_hatch_diagonal(x, y, 3, True)

    def _triple_hatch_right(self, x, y):
        return self._n_hatch_diagonal(x, y, 3, False)

    def _triple_cross_hatch(self, x, y):
        return self._n_hatch_diagonal(x, y, 3, True) + \
               self._n_hatch_diagonal(x, y, 3, False)

    def _n_hatch_diagonal(self, x, y, n, left=False):
        """Hatch pattern generator function."""
        d = 1/float(n)
        result = []
        if left:
            for i in range(n):
                result.append(Line(
                    Point(self.left(x), self.top(y + d * i)),
                    Point(self.right(x - d * i), self.bottom(y))
                ))
                if n:
                    result.append(Line(
                        Point(self.right(x - d * i), self.top(y)),
                        Point(self.right(x), self.top(y + d * i))
                    ))
        else:
            for i in range(n):
                result.append(Line(Point(self.left(x), self.top(y + d * i)), Point(self.left(x + d * i), self.top(y))))
                if n:
                    result.append(Line(Point(self.left(x + d * i), self.bottom(y)), Point(self.right(x), self.top(y + d * i))))
        return result

    def _hatch_v(self, x, y):
        return self._n_hatch_straight(x, y, 1, True)

    def _hatch_h(self, x, y):
        return self._n_hatch_straight(x, y, 1, False)

    def _hv_hatch(self, x, y):
        return self._n_hatch_straight(x, y, 1, True) + \
               self._n_hatch_straight(x, y, 1, False)

    def _double_hatch_v(self, x, y):
        return self._n_hatch_straight(x, y, 2, True)

    def _double_hatch_h(self, x, y):
        return self._n_hatch_straight(x, y, 2, False)

    def _double_hv_hatch(self, x, y):
        return self._n_hatch_straight(x, y, 2, True) + \
               self._n_hatch_straight(x, y, 2, False)

    def _triple_hatch_v(self, x, y):
        return self._n_hatch_straight(x, y, 3, True)

    def _triple_hatch_h(self, x, y):
        return self._n_hatch_straight(x, y, 3, False)

    def _triple_hv_hatch(self, x, y):
        return self._n_hatch_straight(x, y, 3, True) + \
               self._n_hatch_straight(x, y, 3, False)

    def _n_hatch_straight(self, x, y, n, vertical=False):
        """Hatch pattern generator function."""
        d = 1 / float(n)
        offset = 1.0 / (n + 1)
        result = []
        if vertical:
            for i in range(n):
                i = i + offset
                result.append(Line(
                    Point(self.left(x + d * i), self.top(y)),
                    Point(self.left(x + d * i), self.bottom(y))
                ))
                #~ if n:
                    #~ result.append(Line(Point(self.right(x-d*i), self.top(y)), Point(self.right(x), self.top(y+d*i))))
        else:
            for i in range(n):
                i = i + offset
                result.append(Line(
                    Point(self.left(x), self.top(y + d * i)),
                    Point(self.right(x), self.top(y + d * i))
                ))
                #~ if n:
                    #~ result.append(Line(Point(self.left(x+d*i), self.bottom(y)), Point(self.right(x), self.top(y+d*i))))
        return result

    def _fill_trail(self, x, y):
        return [
            Line(
                Point(self.left(x + 0.707), self.top(y)),
                Point(self.right(x), self.bottom(y - 0.707))
            ),
            Line(
                Point(self.left(x), self.top(y + 0.707)),
                Point(self.right(x - 0.707), self.bottom(y))
            )
        ]

    def _fill_foreground(self, x, y):
        return [
            Rectangle(
                Point(self.left(x), self.top(y)),
                Point(self.right(x), self.bottom(y))
            )
        ]

    def _fill_background(self, x, y):
        return []

    def _fill_small_circle(self, x, y):
        return [
            Circle(Point(self.left(x + 0.5), self.top(y + 0.5)), 0.2)
        ]

    def _fill_medium_circle(self, x, y):
        return [
            Circle(Point(self.left(x + 0.5), self.top(y + 0.5)), 0.4)
        ]

    def _fill_large_circle(self, x, y):
        return [
            Circle(Point(self.left(x + 0.5), self.top(y + 0.5)), 0.9)
        ]

    def _fill_qmark(self, x, y):
        return [
            Label(Point(self.left(x), self.bottom(y)), '?')
        ]

    def _fill_triangles(self, x, y):
        return [
            Line(Point(self.left(x + 0.5), self.top(y + 0.2)), Point(self.left(x + 0.75), self.top(y + 0.807))),
            Line(Point(self.left(x + 0.7), self.top(y + 0.807)), Point(self.left(x + 0.25), self.top(y + 0.807))),
            Line(Point(self.left(x + 0.25), self.top(y + 0.807)), Point(self.left(x + 0.5), self.top(y + 0.2))),
        ]

    FILL_TYPES = [
        ('A', '_hatch_left'),
        ('B', '_hatch_right'),
        ('C', '_cross_hatch'),
        ('D', '_double_hatch_left'),
        ('E', '_double_hatch_right'),
        ('F', '_double_cross_hatch'),
        ('G', '_triple_hatch_left'),
        ('H', '_triple_hatch_right'),
        ('I', '_triple_cross_hatch'),
        ('J', '_hatch_v'),
        ('K', '_hatch_h'),
        ('L', '_hv_hatch'),
        ('M', '_double_hatch_v'),
        ('N', '_double_hatch_h'),
        ('O', '_double_hv_hatch'),
        ('P', '_triple_hatch_v'),
        ('Q', '_triple_hatch_h'),
        ('R', '_triple_hv_hatch'),
        ('S', '_fill_qmark'),
        ('T', '_fill_trail'),
        ('U', '_fill_small_circle'),
        ('V', '_fill_medium_circle'),
        ('W', '_fill_large_circle'),
        ('X', '_fill_foreground'),
        ('Y', '_fill_triangles'),
        ('Z', '_fill_background'),
    ]

    FILL_CHARACTERS = ''.join([t+t.lower() for (t, f) in FILL_TYPES])

    def get_fill(self, character):
        """Return fill function based on character."""
        for head, function_name in self.FILL_TYPES:
            if character == head:
                return getattr(self, function_name)
        raise ValueError('no such fill type: {!r}'.format(character))

    # - - - - - - - - - fixed characters and their shapes - - - - - - - - -

    def _open_triangle_left(self, x, y):
        return [
            Line(
                Point(self.left(x), self.vcenter(y)),
                Point(self.right(x), self.top(y))
            ),
            Line(
                Point(self.left(x), self.vcenter(y)),
                Point(self.right(x), self.bottom(y))
            )
        ]

    def _open_triangle_right(self, x, y):
        return [
            Line(
                Point(self.right(x), self.vcenter(y)),
                Point(self.left(x), self.top(y))
            ),
            Line(
                Point(self.right(x), self.vcenter(y)),
                Point(self.left(x), self.bottom(y))
            )
        ]

    def _circle(self, x, y):
        return [
            Circle(Point(self.hcenter(x), self.vcenter(y)), NOMINAL_SIZE / 2.0)
        ]

    FIXED_TYPES = [
        ('{', '_open_triangle_left'),
        ('}', '_open_triangle_right'),
        ('*', '_circle'),
    ]

    FIXED_CHARACTERS = ''.join([t for (t, f) in FIXED_TYPES])

    def get_fixed_character(self, character):
        """Return fixed character function."""
        for head, function_name in self.FIXED_TYPES:
            if character == head:
                return getattr(self, function_name)
        raise ValueError('no such character: {!r}'.format(character))

    # - - - - - - - - - helper function for shape recognition - - - - - - - - -

    def _follow_vertical_line(self, x, y):
        """Find a vertical line with optional arrow heads."""
        # follow line to the bottom
        _, end_y, line_end_style = self._follow_line(x, y, dy=1, line_character='|')
        # follow line to the top
        _, start_y, line_start_style = self._follow_line(x, y, dy=-1, line_character='|')
        # if a '+' follows a line, then the line is stretched to hit the '+' center
        start_y_fix = end_y_fix = 0
        if self.get(x, start_y - 1) == '+':
            start_y_fix = -0.5
        if self.get(x, end_y + 1) == '+':
            end_y_fix = 0.5
        # tag characters as used (not the arrow heads)
        self.tag([(x, y) for y in range(start_y, end_y + 1)], CLASS_LINE)
        # return the new shape object with arrows etc.
        p1 = complex(self.hcenter(x), self.top(start_y + start_y_fix))
        p2 = complex(self.hcenter(x), self.bottom(end_y + end_y_fix))
        shapes = []
        if line_start_style:
            p1, arrow_shapes = line_start_style(p1, p2)
            shapes.extend(arrow_shapes)
        if line_end_style:
            p2, arrow_shapes = line_end_style(p2, p1)
            shapes.extend(arrow_shapes)
        shapes.append(Line(p1, p2))
        return group(shapes)

    def _follow_horizontal_line(self, x, y, thick=False):
        """Find a horizontal line with optional arrow heads."""
        if thick:
            line_character = '='
        else:
            line_character = '-'
        # follow line to the right
        end_x, _, line_end_style = self._follow_line(x, y, dx=1, line_character=line_character)
        # follow line to the left
        start_x, _, line_start_style = self._follow_line(x, y, dx=-1, line_character=line_character)
        start_x_fix = end_x_fix = 0
        if self.get(start_x - 1, y) == '+':
            start_x_fix = -0.5
        if self.get(end_x + 1, y) == '+':
            end_x_fix = 0.5
        self.tag([(x, y) for x in range(start_x, end_x+1)], CLASS_LINE)
        # return the new shape object with arrows etc.
        p1 = complex(self.left(start_x + start_x_fix), self.vcenter(y))
        p2 = complex(self.right(end_x + end_x_fix), self.vcenter(y))
        shapes = []
        if line_start_style:
            p1, arrow_shapes = line_start_style(p1, p2)
            shapes.extend(arrow_shapes)
        if line_end_style:
            p2, arrow_shapes = line_end_style(p2, p1)
            shapes.extend(arrow_shapes)
        shapes.append(Line(p1, p2, thick=thick))
        return group(shapes)

    def _follow_lower_horizontal_line(self, x, y):
        """\
        Find a horizontal line, the line is aligned to the bottom and a bit
        wider, so that it can be used for shapes like this:
           ___
        __|   |___
        """
        # follow line to the right
        end_x, _, line_end_style = self._follow_line(x, y, dx=1, line_character='_', arrows=False)
        # follow line to the left
        start_x, _, line_start_style = self._follow_line(x, y, dx=-1, line_character='_', arrows=False)
        self.tag([(x, y) for x in range(start_x, end_x + 1)], CLASS_LINE)
        # return the new shape object with arrows etc.
        p1 = complex(self.hcenter(start_x - 1), self.bottom(y))
        p2 = complex(self.hcenter(end_x + 1), self.bottom(y))
        return [Line(p1, p2)]

    def _follow_upper_horizontal_line(self, x, y):
        """\
        Find a horizontal line, the line is aligned to the bottom and a bit
        wider, so that it can be used for shapes like this:

          |~~~|
        ~~     ~~~
        """
        # follow line to the right
        end_x, _, line_end_style = self._follow_line(x, y, dx=1, line_character='~', arrows=False)
        # follow line to the left
        start_x, _, line_start_style = self._follow_line(x, y, dx=-1, line_character='~', arrows=False)
        self.tag([(x, y) for x in range(start_x, end_x + 1)], CLASS_LINE)
        # return the new shape object with arrows etc.
        p1 = complex(self.hcenter(start_x - 1), self.top(y))
        p2 = complex(self.hcenter(end_x + 1), self.top(y))
        return [Line(p1, p2)]

    def _follow_line(self, x, y, dx=0, dy=0, line_character=None, arrows=True):
        """Helper function for all the line functions."""
        # follow line in the given direction
        while 0 <= x < self.width and 0 <= y < self.height and self.get(x+dx, y+dy) == line_character:
            x += dx
            y += dy
        if arrows:
            # check for arrow head
            following_character = self.get(x + dx, y + dy)
            if following_character in self.ARROW_HEADS:
                line_end_style = self.get_arrow(following_character, dx, dy)
                if line_end_style:
                    x += dx
                    y += dy
            else:
                line_end_style = None
        else:
            line_end_style = None
        return x, y, line_end_style

    def _plus_joiner(self, x, y):
        """\
        Adjacent '+' signs are connected with a line from center to center
        required for images like these:

           +---+         The box should be closed on all sides
           |   +--->     and the arrow start should touch the box.
           +---+
        """
        result = []
        #~ for dx, dy in ((1,0), (-1,0), (0,1), (0,-1)):
        # looking right and down is sufficient as the scan is done from left to
        # right, top to bottom
        for dx, dy in ((1, 0), (0, 1)):
            if self.get(x + dx, y + dy) == '+':
                result.append(Line(
                    Point(self.hcenter(x), self.vcenter(y)),
                    Point(self.hcenter(x + dx), self.vcenter(y + dy))
                ))
        self.tag([(x, y)], CLASS_JOIN)
        return result

    def _follow_fill(self, character, start_x, start_y):
        """\
        Fill shapes like the ones below with a pattern. when the character is
        upper case, draw a border too.

            XXX  aaa  BB
           XXX    a
        """
        fill = self.get_fill(character.upper())
        border = character.isupper()
        result = []
        # flood fill algorithm, searching for similar characters
        to_scan = [(start_x, start_y)]
        while to_scan:
            x, y = to_scan.pop()
            if self.cls(x, y) is None:
                if self.get(x, y) == character:
                    result.extend(fill(x, y))
                    self.tag([(x, y)], CLASS_RECTANGLE)
                if self.get(x + 1, y) == character:
                    if self.cls(x + 1, y) is None:
                        to_scan.append((x + 1, y))
                elif border:
                    result.append(Line(
                        Point(self.right(x), self.top(y)),
                        Point(self.right(x), self.bottom(y))))
                if self.get(x - 1, y) == character:
                    if self.cls(x - 1, y) is None:
                        to_scan.append((x - 1, y))
                elif border:
                    result.append(Line(
                        Point(self.left(x), self.top(y)),
                        Point(self.left(x), self.bottom(y))))
                if self.get(x, y + 1) == character:
                    if self.cls(x, y + 1) is None:
                        to_scan.append((x, y + 1))
                elif border:
                    result.append(Line(
                        Point(self.left(x), self.bottom(y)),
                        Point(self.right(x), self.bottom(y))))
                if self.get(x, y - 1) == character:
                    if self.cls(x, y - 1) is None:
                        to_scan.append((x, y - 1))
                elif border:
                    result.append(Line(
                        Point(self.left(x), self.top(y)),
                        Point(self.right(x), self.top(y))))
        return group(result)

    def _follow_horizontal_string(self, start_x, y, accept_anything=False, quoted=False):
        """\
        Find a string. may contain single spaces, but the detection is
        aborted after more than one space.

              Text one   "Text two"

        accept_anything means that all non space characters are interpreted
        as text.
        """
        # follow line from left to right
        if quoted:
            quotation_character = self.get(start_x, y)
            x = start_x + 1
        else:
            quotation_character = None
            x = start_x
        text = []
        if self.get(x, y) != ' ':
            text.append(self.get(x, y))
            self.tag([(x, y)], CLASS_STRING)
            is_first_space = True
            while 0 <= x + 1 < self.width and self.cls(x + 1, y) is None:
                if not quoted:
                    if self.get(x + 1, y) == ' ' and not is_first_space:
                        break
                    if not accept_anything and not self.get(x + 1, y).isalnum():
                        break
                x += 1
                character = self.get(x, y)
                if character == quotation_character:
                    self.tag([(x, y)], CLASS_STRING)
                    break
                text.append(character)
                if character == ' ':
                    is_first_space = False
                else:
                    is_first_space = True
            if text[-1] == ' ':
                del text[-1]
                x -= 1
            self.tag([(x, y) for x in range(start_x, x + 1)], CLASS_STRING)
            return [Label(
                Point(self.left(start_x), self.bottom(y)),
                ''.join(text)
            )]
        else:
            return []

    def _follow_rounded_edge(self, x, y):
        """\
        Check for rounded edges:
            /-    |     -\-    |   and also \    /  etc.
            |    -/      |     \-            -  |
        """
        result = []
        if self.get(x, y) == '/':
            # rounded rectangles
            if (self.get(x + 1, y) == '-' and self.get(x, y + 1) == '|'):
                # upper left corner
                result.append(Arc(
                    Point(self.hcenter(x), self.bottom(y)), 90,
                    Point(self.right(x), self.vcenter(y)),  180
                ))
            if self.get(x - 1, y) == '-' and self.get(x, y - 1) == '|':
                # lower right corner
                result.append(Arc(
                    Point(self.hcenter(x), self.top(y)), -90,
                    Point(self.left(x), self.vcenter(y)), 0
                ))
            if not result:
                # if used as diagonal line
                p1 = p2 = None
                a1 = a2 = 0
                arc = c1 = c2 = False
                if self.get(x + 1, y - 1) == '|':
                    p1 = Point(self.hcenter(x + 1), self.top(y))
                    a1 = -90
                    arc = c1 = True
                elif self.get(x + 1, y - 1) == '+':
                    p1 = Point(self.hcenter(x + 1), self.vcenter(y - 1))
                    a1 = -135
                elif self.get(x + 1, y - 1) == '-':
                    p1 = Point(self.right(x), self.vcenter(y - 1))
                    a1 = 180
                    arc = c1 = True
                elif self.get(x + 1, y - 1) == '/':
                    p1 = Point(self.right(x), self.top(y))
                    a1 = -135
                    c1 = True
                elif self.get(x + 1, y) == '|':
                    p1 = Point(self.hcenter(x + 1), self.top(y))
                elif self.get(x, y - 1) == '-':
                    p1 = Point(self.right(x), self.vcenter(y - 1))

                if self.get(x - 1, y + 1) == '|':
                    p2 = Point(self.hcenter(x - 1), self.top(y + 1))
                    a2 = 90
                    arc = c2 = True
                elif self.get(x - 1, y + 1) == '+':
                    p2 = Point(self.hcenter(x - 1), self.vcenter(y + 1))
                    a2 = 45
                elif self.get(x - 1, y + 1) == '-':
                    p2 = Point(self.left(x), self.vcenter(y + 1))
                    a2 = 0
                    arc = c2 = True
                elif self.get(x - 1, y + 1) == '/':
                    p2 = Point(self.left(x), self.bottom(y))
                    a2 = 45
                    c2 = True
                elif self.get(x - 1, y) == '|':
                    p2 = Point(self.hcenter(x - 1), self.bottom(y))
                elif self.get(x, y + 1) == '-':
                    p2 = Point(self.left(x), self.vcenter(y + 1))

                if p1 or p2:
                    if not p1:
                        p1 = Point(self.right(x), self.top(y))
                    if not p2:
                        p2 = Point(self.left(x), self.bottom(y))
                    if arc and self.rounded:
                        result.append(Arc(p1, a1, p2, a2, c1, c2))
                    else:
                        result.append(Line(p1, p2))
        else:  # '\'
            # rounded rectangles
            if self.get(x-1, y) == '-' and self.get(x, y + 1) == '|':
                # upper right corner
                result.append(Arc(
                    Point(self.hcenter(x), self.bottom(y)), 90,
                    Point(self.left(x), self.vcenter(y)), 0
                ))
            if self.get(x+1, y) == '-' and self.get(x, y - 1) == '|':
                # lower left corner
                result.append(Arc(
                    Point(self.hcenter(x), self.top(y)), -90,
                    Point(self.right(x), self.vcenter(y)), 180
                ))
            if not result:
                # if used as diagonal line
                p1 = p2 = None
                a1 = a2 = 0
                arc = c1 = c2 = False
                if self.get(x - 1, y - 1) == '|':
                    p1 = Point(self.hcenter(x-1), self.top(y))
                    a1 = -90
                    arc = c1 = True
                elif self.get(x - 1, y - 1) == '+':
                    p1 = Point(self.hcenter(x-1), self.vcenter(y - 1))
                    a1 = -45
                elif self.get(x - 1, y - 1) == '-':
                    p1 = Point(self.left(x), self.vcenter(y-1))
                    a1 = 0
                    arc = c1 = True
                elif self.get(x - 1, y - 1) == '\\':
                    p1 = Point(self.left(x), self.top(y))
                    a1 = -45
                    c1 = True
                elif self.get(x - 1, y) == '|':
                    p1 = Point(self.hcenter(x-1), self.top(y))
                elif self.get(x, y - 1) == '-':
                    p1 = Point(self.left(x), self.hcenter(y - 1))

                if self.get(x + 1, y + 1) == '|':
                    p2 = Point(self.hcenter(x+1), self.top(y + 1))
                    a2 = 90
                    arc = c2 = True
                elif self.get(x + 1, y + 1) == '+':
                    p2 = Point(self.hcenter(x+1), self.vcenter(y + 1))
                    a2 = 135
                elif self.get(x + 1, y + 1) == '-':
                    p2 = Point(self.right(x), self.vcenter(y + 1))
                    a2 = 180
                    arc = c2 = True
                elif self.get(x + 1, y + 1) == '\\':
                    p2 = Point(self.right(x), self.bottom(y))
                    a2 = 135
                    c2 = True
                elif self.get(x + 1, y) == '|':
                    p2 = Point(self.hcenter(x+1), self.bottom(y))
                elif self.get(x, y + 1) == '-':
                    p2 = Point(self.right(x), self.vcenter(y + 1))

                if p1 or p2:
                    if not p1:
                        p1 = Point(self.left(x), self.top(y))
                    if not p2:
                        p2 = Point(self.right(x), self.bottom(y))
                    if arc and self.rounded:
                        result.append(Arc(p1, a1, p2, a2, c1, c2))
                    else:
                        result.append(Line(p1, p2))
        if result:
            self.tag([(x, y)], CLASS_JOIN)
        return group(result)


def process(input, visitor_class, options=None):
    """\
    Parse input and render using the given visitor class.

    :param input: String or file like object with the image as text.

    :param visitor_class: A class object, it will be used to render the
        resulting image.

    :param options: A dictionary containing the settings. When ``None`` is
        given, defaults are used.

    :returns: instantiated ``visitor_class`` and the image has already been
        processed with the visitor.

    :exception: This function can raise an ``UnsupportedFormatError`` exception
        if the specified format is not supported.
    """

    # remember user options (don't want to rename function parameter above)
    user_options = options
    # start with a copy of the defaults
    options = DEFAULT_OPTIONS.copy()
    if user_options is not None:
        # override with settings passed by caller
        options.update(user_options)

    if 'fill' not in options or options['fill'] is None:
        options['fill'] = options['foreground']

    # if input is a file like object, read from it (otherwise it is assumed to
    # be a string)
    if hasattr(input, 'read'):
        input = input.read()

    if options['debug']:
        sys.stderr.write('{!r}\n'.format(input))

    aaimg = AsciiArtImage(input, options)

    if options['debug']:
        sys.stderr.write('{}\n'.format(aaimg))
    aaimg.recognize()

    visitor = visitor_class(options)
    visitor.visit_image(aaimg)
    return visitor


def render(input, output=None, options=None):
    """\
    Render an ASCII art figure to a file or file-like.

    :param input: If ``input`` is a basestring subclass (str or unicode), the
        text contained in ``input`` is rendered. If ``input is a file-like
        object, the text to render is taken using ``input.read()``.

    :param output: If no ``output`` is specified, the resulting rendered image
        is returned as a string. If output is a basestring subclass, a file
        with the name of ``output`` contents is created and the rendered image
        is saved there. If ``output`` is a file-like object, ``output.write()``
        is used to save the rendered image.

    :param options: A dictionary containing the settings. When ``None`` is
        given, defaults are used.

    :returns: This function returns a tuple ``(visitor, output)``, where
        ``visitor`` is visitor instance that rendered the image and ``output``
        is the image as requested by the ``output`` parameter (a ``str`` if it
        was ``None``, or a file-like object otherwise, which you should
        ``close()`` if needed).

    :exception: This function can raise an ``UnsupportedFormatError`` exception
        if the specified format is not supported.
    """

    if options is None:
        options = {}

    close_output = False
    if output is None:
        from io import BytesIO
        options['file_like'] = BytesIO()
    elif isinstance(output, basestring):
        options['file_like'] = open(output, 'wb')
        close_output = True
    else:
        options['file_like'] = output
    try:
        # late import of visitor classes to not cause any import errors for
        # unsupported backends (this would happen when a library a backend
        # depends on is not installed)
        if options['format'].lower() == 'svg':
            from . import svg
            visitor_class = svg.SVGOutputVisitor
        elif options['format'].lower() == 'pdf':
            from . import pdf
            visitor_class = pdf.PDFOutputVisitor
        elif options['format'].lower() == 'ascii':
            from . import aa
            visitor_class = aa.AsciiOutputVisitor
        else:
            # for all other formats, it may be a bitmap type. let
            # PIL decide if it can write a file of that type.
            from . import pil
            visitor_class = pil.PILOutputVisitor
        # now render and output the image
        visitor = process(input, visitor_class, options)
    finally:
        if close_output:
            options['file_like'].close()
    file_like = options.pop('file_like')
    return (visitor, file_like)


def main():
    """implement an useful main for use as command line program"""
    import sys
    import optparse
    import os.path

    parser = optparse.OptionParser(
        usage="%prog [options] [file]",
        version="""\
%prog 0.5

Copyright (C) 2006-2017 aafigure-team

Redistribution and use in source and binary forms, with or without
modification, are permitted under the terms of the BSD License.

THIS SOFTWARE IS PROVIDED BY THE AAFIGURE-TEAM ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AAFIGURE-TEAM BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
""",
        description="ASCII art to image (SVG, PNG, JPEG, PDF and more) converter."
    )

    parser.add_option(
        "-e", "--encoding",
        dest="encoding",
        action="store",
        help="character encoding of input text",
        default=DEFAULT_OPTIONS['encoding'],
    )

    parser.add_option(
        "-w", "--wide-chars",
        dest="widechars",
        action="store",
        help="unicode properties to be treated as wide glyph (e.g. 'F,W,A')",
        default=DEFAULT_OPTIONS['widechars'],
    )

    parser.add_option(
        "-o", "--output",
        dest="output",
        metavar="FILE",
        help="write output to FILE"
    )

    parser.add_option(
        "-t", "--type",
        dest="format",
        help="filetype: png, jpg, svg (by default autodetect from filename)",
        default=None,
    )

    parser.add_option(
        "-D", "--debug",
        dest="debug",
        action="store_true",
        help="enable debug outputs",
        default=DEFAULT_OPTIONS['debug'],
    )

    parser.add_option(
        "-T", "--textual",
        dest="textual",
        action="store_true",
        help="disable horizontal fill detection",
        default=DEFAULT_OPTIONS['textual'],
    )

    parser.add_option(
        "-S", "--textual-strict",
        dest="textual_strict",
        action="store_true",
        help="disable horizontal and vertical fill detection",
        default=DEFAULT_OPTIONS['textual_strict'],
    )

    parser.add_option(
        "-s", "--scale",
        dest="scale",
        action="store",
        type='float',
        help="set scale",
        default=DEFAULT_OPTIONS['scale'],
    )

    parser.add_option(
        "-a", "--aspect",
        dest="aspect",
        action="store",
        type='float',
        help="set aspect ratio",
        default=DEFAULT_OPTIONS['aspect'],
    )

    parser.add_option(
        "-l", "--linewidth",
        dest="line_width",
        action="store",
        type='float',
        help="set width, svg only",
        default=DEFAULT_OPTIONS['line_width'],
    )

    parser.add_option(
        "--proportional",
        dest="proportional",
        action="store_true",
        help="use proportional font instead of fixed width",
        default=DEFAULT_OPTIONS['proportional'],
    )

    parser.add_option(
        "-f", "--foreground",
        dest="foreground",
        action="store",
        help="foreground color default=%default",
        default=DEFAULT_OPTIONS['foreground'],
    )

    parser.add_option(
        "-x", "--fill",
        dest="fill",
        action="store",
        help="foreground color default=foreground",
        default=None,
    )

    parser.add_option(
        "-b", "--background",
        dest="background",
        action="store",
        help="foreground color default=%default",
        default=DEFAULT_OPTIONS['background'],
    )

    parser.add_option(
        "--rounded",
        dest="rounded",
        action="store_true",
        help="use arcs for rounded edges instead of straight lines",
        default=DEFAULT_OPTIONS['proportional'],
    )

    parser.add_option(
        "-O", "--option",
        dest="_extra_options",
        action="append",
        help="pass special options to backends (expert user)",
    )

    (options, args) = parser.parse_args()

    if len(args) > 1:
        parser.error("too many arguments")

    if options.format is None:
        if options.output is None:
            parser.error("Please specify output format with --type")
        else:
            options.format = os.path.splitext(options.output)[1][1:]

    if args:
        input = open(args[0])
    else:
        input = sys.stdin
    #~ input = codecs.getreader(options.encoding)(input)

    if options.output is None:
        output = sys.stdout
    else:
        output = open(options.output, 'wb')

    # explicit copying of parameters to the options dictionary
    options_dict = {}
    for key in ('widechars', 'textual', 'textual_strict',
                'proportional', 'line_width', 'aspect',
                'scale', 'format', 'debug'):
        options_dict[key] = getattr(options, key)
    # ensure all color parameters start with a '#'
    # this is for the convenience of the user as typing the shell comment
    # character isn't for everyone ;-)
    for color in ('foreground', 'background', 'fill'):
        value = getattr(options, color)
        if value is not None:
            if value[0] != '#':
                options_dict[color] = '#{}'.format(value)
            else:
                options_dict[color] = value
    # copy extra options
    if options._extra_options is not None:
        for keyvalue in options._extra_options:
            try:
                key, value = keyvalue.split('=')
            except ValueError:
                parser.error('--option must be in the format <key>=<value> (not {!r})'.format(keyvalue))
            options_dict[key] = value

    if options.debug:
        sys.stderr.write('options={!r}\n'.format(options_dict))

    try:
        (visitor, output) = render(input, output, options_dict)
        output.close()
    except UnsupportedFormatError as e:
        print("ERROR: Can't output format '{}': {}".format(options.format, e))


# when module is run, run the command line tool
if __name__ == '__main__':
    main()



#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2007 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""
AAFigure package.
"""

from .aafigure import process, render, UnsupportedFormatError, AsciiArtImage



#!/usr/bin/env python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2017 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause

# command line stub for aafigure

from . import aafigure
aafigure.main()



#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2008 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
PDF renderer for the aafigure package.
"""

import sys
from .error import UnsupportedFormatError
try:
    # import reportlab
    from reportlab.lib import colors
    from reportlab.graphics.shapes import *
    from reportlab.graphics import renderPDF
    from reportlab.pdfbase import pdfmetrics
    from reportlab.pdfbase.ttfonts import TTFont
except ImportError:
    raise UnsupportedFormatError('please install Reportlab to get PDF output support')


class PDFOutputVisitor:
    """Render a list of shapes as PDF vector image."""

    def __init__(self, options):
        """\
        Dual use as PDF file writer or as Reportlab Drawing generator.

        files: file_like or filname is given in the options:
            The PDF file is written there.

        Drawing: file_like and filname are missing in the options:
            No output is generated. The Drawing can be used for example::

                visitor = PDFOutputVisitor(None, ...)
                do_something(renderPDF.GraphicsFlowable(visitor.drawing))
        """
        self.options = options
        self.scale = 4 * options['scale']
        self.line_width = 0.4 * options['line_width']
        self.foreground = options['foreground']
        self.background = options['background']
        self.fillcolor = options['fill']
        # if front is given explicit, use it instead of textual/proportional flags
        if 'font' in options:
            self.font = options['font']
            if self.font.endswith('.ttf'):
                # ttf support
                pdfmetrics.registerFont(TTFont(self.font, self.font))
        else:
            if options['proportional']:
                self.font = 'Helvetica'
            else:
                self.font = 'Courier'

    def _num(self, number):
        """helper to format numbers with scale for PDF output"""
        return number * self.scale

    def _color(self, color):
        return colors.HexColor(color)

    def visit_image(self, aa_image):
        """
        Process the given ASCIIArtFigure and output the shapes in
        the PDF file
        """
        self.aa_image = aa_image        # save for later XXX not optimal to do it here
        self.width = aa_image.width * aa_image.nominal_size * aa_image.aspect_ratio
        self.height = aa_image.height * aa_image.nominal_size
        self.drawing = Drawing(self._num(self.width), self._num(self.height))
        self.visit_shapes(aa_image.shapes)
        # if file is given, write
        if 'file_like' in self.options:
            renderPDF.drawToFile(self.drawing, self.options['file_like'], '')

    def visit_shapes(self, shapes):
        for shape in shapes:
            shape_name = shape.__class__.__name__.lower()
            visitor_name = 'visit_{}'.format(shape_name)
            if hasattr(self, visitor_name):
                getattr(self, visitor_name)(shape)
            else:
                sys.stderr.write("WARNING: don't know how to handle shape {!r}\n".format(shape))

    # - - - - - - PDF drawing helpers - - - - - - -
    def _line(self, x1, y1, x2, y2, thick):
        """Draw a line, coordinates given as four decimal numbers"""
        self.drawing.add(Line(
            self._num(x1), self._num(self.height - y1),
            self._num(x2), self._num(self.height - y2),
            strokeColor=self._color(self.foreground),
            strokeWidth=self.line_width*(1 + 0.5 * bool(thick))))

    def _rectangle(self, x1, y1, x2, y2, style=''):
        """Draw a rectangle, coordinates given as four decimal numbers."""
        if x1 > x2:
            x1, x2 = x2, x1
        if y1 > y2:
            y1, y2 = y2, y1
        self.drawing.add(Rect(
            self._num(x1), self._num(self.height - y2),
            self._num(x2 - x1), self._num(y2 - y1),
            fillColor=self._color(self.fillcolor),
            strokeWidth=self.line_width))

    # - - - - - - visitor function for the different shape types - - - - - - -

    def visit_point(self, point):
        self.drawing.add(Circle(
            self._num(point.x), self._num(self.height - point.y),
            self._num(0.2),
            fillColor=self._color(self.foreground),
            strokeWidth=self.line_width))

    def visit_line(self, line):
        x1, x2 = line.start.x, line.end.x
        y1, y2 = line.start.y, line.end.y
        self._line(x1, y1, x2, y2, line.thick)

    def visit_rectangle(self, rectangle):
        self._rectangle(
            rectangle.p1.x, rectangle.p1.y,
            rectangle.p2.x, rectangle.p2.y)

    def visit_circle(self, circle):
        self.drawing.add(Circle(
            self._num(circle.center.x), self._num(self.height - circle.center.y),
            self._num(circle.radius),
            strokeColor=self._color(self.foreground),
            fillColor=self._color(self.fillcolor),
            strokeWidth=self.line_width))

    def visit_label(self, label):
        #  font-weight="bold"   style="stroke:%s"
        self.drawing.add(String(
            self._num(label.position.x), self._num(self.height - label.position.y + self.aa_image.nominal_size * 0.2),
            label.text,
            fontSize=self._num(self.aa_image.nominal_size),
            fontName=self.font,
            fillColor=self._color(self.foreground)))

    def visit_group(self, group):
        # XXX could add a group to the PDF file
        self.visit_shapes(group.shapes)

    def visit_arc(self, arc):
        p1, p2 = arc.start, arc.end
        c1 = arc.start_control_point()
        c2 = arc.end_control_point()
        path = Path(strokeColor=self._color(self.foreground),
                    strokeWidth=self.line_width)
        path.moveTo(self._num(p1.x), self._num(self.height - p1.y))
        path.curveTo(self._num(c1.x), self._num(self.height - c1.y),
                     self._num(c2.x), self._num(self.height - c2.y),
                     self._num(p2.x), self._num(self.height - p2.y))
        self.drawing.add(path)



#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2006 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
SVG renderer for the aafigure package.
"""

import sys
import codecs
from xml.sax.saxutils import escape


class SVGOutputVisitor:
    """Render a list of shapes as SVG image."""

    def __init__(self, options):
        self.options = options
        self.file_like = codecs.getwriter('utf-8')(options['file_like'])
        self.scale = options['scale'] * 7
        self.line_width = options['line_width']
        self.foreground = options['foreground']
        self.background = options['background']
        self.fillcolor = options['fill']
        self.border = 3
        self.indent = ''
        # if front is given explicit, use it instead of textual/proportional flags
        if 'font' in options:
            self.font = options['font']
        else:
            if options['proportional']:
                self.font = u'sans-serif'
            else:
                self.font = u'monospace'

    def _num(self, number):
        """helper to scale numbers for svg output"""
        return number * self.scale

    def _coordinate(self, number):
        """helper to scale numbers for svg output"""
        return self._num(number) + self.border

    def get_size_attrs(self):
        """get image size as svg text"""
        # this function is here beacuse of a hack. the rst2html converter
        # has to know the size of the figure it inserts
        return u'width="{}" height="{}"'.format(
            self._num(self.width) + 2 * self.border,
            self._num(self.height) + 2 * self.border)

    def visit_image(self, aa_image, xml_header=True):
        """\
        Process the given ASCIIArtFigure and output the shapes in
        the SVG file
        """
        self.aa_image = aa_image        # save for later XXX not optimal to do it here
        self.width = aa_image.width * aa_image.nominal_size * aa_image.aspect_ratio
        self.height = aa_image.height * aa_image.nominal_size
        if xml_header:
            self.file_like.write(
                u'<?xml version="1.0" standalone="no"?>\n'
                '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" '
                '"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n\n'
                '<svg width="{w}" height="{h}" viewBox="0 0 {w} {h}" version="1.1" xmlns="http://www.w3.org/2000/svg" '
                'xmlns:xlink="http://www.w3.org/1999/xlink">\n'
                '<!-- automatically generated by aafigure -->'.format(
                    w=self._num(self.width) + 2 * self.border,
                    h=self._num(self.height) + 2 * self.border))
        else:
            self.file_like.write(
                u'<svg width="{w}" height="{h}" viewBox="0 0 {w} {h}" version="1.1" '
                'xmlns="http://www.w3.org/2000/svg">\n'.format(
                    w=self._num(self.width) + 2 * self.border,
                    h=self._num(self.height) + 2 * self.border))
        self.visit_shapes(aa_image.shapes)
        self.file_like.write(u'</svg>\n')

    def visit_shapes(self, shapes):
        for shape in shapes:
            shape_name = shape.__class__.__name__.lower()
            visitor_name = 'visit_{}'.format(shape_name)
            if hasattr(self, visitor_name):
                getattr(self, visitor_name)(shape)
            else:
                sys.stderr.write(u"WARNING: don't know how to handle shape {!r}\n".format(shape))

    # - - - - - - SVG drawing helpers - - - - - - -
    def _line(self, x1, y1, x2, y2, thick):
        """Draw a line, coordinates given as four decimal numbers"""
        self.file_like.write(
            u'{}<line x1="{}" y1="{}" x2="{}" y2="{}" stroke="{}" '
            'stroke-width="{}" />\n'.format(
                self.indent,
                self._coordinate(x1),
                self._coordinate(y1),
                self._coordinate(x2),
                self._coordinate(y2),
                self.foreground,
                self.line_width * (1 + bool(thick))))

    def _rectangle(self, x1, y1, x2, y2, style=''):
        """\
        Draw a rectangle, coordinates given as four decimal numbers.
        ``style`` is inserted in the SVG. It could be e.g. "fill:yellow"
        """
        if x1 > x2:
            x1, x2 = x2, x1
        if y1 > y2:
            y1, y2 = y2, y1
        self.file_like.write(
            u'{}<rect x="{}" y="{}" width="{}" height="{}" stroke="{}" '
            'fill="{}" stroke-width="{}" style="{}" />'.format(
                self.indent,
                self._coordinate(x1), self._coordinate(y1),
                self._num(x2 - x1), self._num(y2 - y1),
                self.fillcolor,  # stroke:%s;
                self.fillcolor,
                self.line_width,
                style))

    # - - - - - - visitor function for the different shape types - - - - - - -

    def visit_point(self, point):
        self.file_like.write(
            u'{}<circle cx="{}" cy="{}" r="{}" fill="{}" stroke="{}" '
            'stroke-width="{}" />'.format(
                self.indent,
                self._coordinate(point.x), self._coordinate(point.y),
                self._num(0.2),
                self.foreground, self.foreground,
                self.line_width))

    def visit_line(self, line):
        x1, x2 = line.start.x, line.end.x
        y1, y2 = line.start.y, line.end.y
        self._line(x1, y1, x2, y2, line.thick)

    def visit_rectangle(self, rectangle):
        self._rectangle(
            rectangle.p1.x, rectangle.p1.y,
            rectangle.p2.x, rectangle.p2.y)

    def visit_circle(self, circle):
        self.file_like.write(
            u'{}<circle cx="{}" cy="{}" r="{}" stroke="{}" stroke-width="{}" '
            'fill="{}" />'.format(
                self.indent,
                self._coordinate(circle.center.x), self._coordinate(circle.center.y),
                self._num(circle.radius),
                self.foreground,
                self.line_width,
                self.fillcolor))

    def visit_label(self, label):
        #  font-weight="bold"   style="stroke:%s"
        self.file_like.write(
            u'{}<text x="{}" y="{}" font-family="{}" font-size="{}" '
            'fill="{}" >\n  {}\n{}</text>\n'.format(
                self.indent,
                self._coordinate(label.position.x), self._coordinate(label.position.y - 0.3),  # XXX static offset not good in all situations
                self.font,
                self._num(self.aa_image.nominal_size),
                self.foreground,
                escape(label.text),
                self.indent))

    def visit_group(self, group):
        self.file_like.write(u'<g>\n')
        old_indent = self.indent
        self.indent += u'    '
        self.visit_shapes(group.shapes)
        self.indent = old_indent
        self.file_like.write(u'</g>\n')

    def visit_arc(self, arc):
        p1, p2 = arc.start, arc.end
        c1 = arc.start_control_point()
        c2 = arc.end_control_point()
        self.file_like.write(
            u'{}<path d="M{},{} C{},{} {},{} {},{}" fill="none" '
            'stroke="{}" stroke-width="{}" />'.format(
                self.indent,
                self._coordinate(p1.x), self._coordinate(p1.y),
                self._coordinate(c1.x), self._coordinate(c1.y),
                self._coordinate(c2.x), self._coordinate(c2.y),
                self._coordinate(p2.x), self._coordinate(p2.y),
                self.foreground,
                self.line_width))



#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2009 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
Exception classes for the aafigure package.
"""

class UnsupportedFormatError(Exception):
    """Error message when an unknown format is specified"""
